/*
 * main.c
 *
 * Update 2016/12/14
 * (C) 2008 Yutaka Hirata
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/completion.h>
#include <linux/kthread.h>

MODULE_LICENSE("Dual BSD/GPL");

static wait_queue_head_t wait;
static struct completion comp;
static struct task_struct *tsk;

static int sample_thread(void *num)
{
	printk("%s called\n", __func__);

	//daemonize("sample_thread");
	allow_signal(SIGTERM);

	for (;;) {
		//sleep_on_timeout(&wait, 3 * HZ);
		schedule_timeout_interruptible(msecs_to_jiffies(3*1000));

		if (signal_pending(current))
			break;
	}

	printk("%s leaved\n", __func__);

	complete_and_exit(&comp, 0);

	return 0;
}

static int sample_init(void)
{
	printk("sample driver installed.\n");

	init_completion(&comp);
	init_waitqueue_head(&wait);

	//thr_pid = kernel_thread(sample_thread, NULL, CLONE_FS | CLONE_FILES);
	tsk = kthread_run(sample_thread, NULL, "sample_thread");
	if (IS_ERR(tsk)) {
		return -EINVAL;
	}
	printk("kernel_thread %d PID\n", tsk->pid);

	return 0;
}

static void sample_exit(void)
{
	printk("sample driver removed.\n");

	/* kill_proc() は削除されたため、カーネルスレッドにシグナルを
	 * 送ることはできなくなった。そもそも、kernel_thread() 自体が
	 * レガシーであったため、Linux 3.0では x86 のコードに残存しているが、
	 * 将来的に完全撤廃される。
	 */
	//kill_proc(thr_pid, SIGTERM, 1);
	send_sig(SIGTERM, tsk, 1);
	wait_for_completion(&comp);
}

module_init(sample_init);
module_exit(sample_exit);

