/*
 * Linux device driver programming Book(SB Creative corp)
 * Procfs sample program
 *
 * Update 2016/12/11 
 * (C) 2008 Yutaka Hirata
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <net/net_namespace.h>

MODULE_LICENSE("Dual BSD/GPL");

/*
 * すでに /proc/driver/ は存在するが、その配下にサブディレクトリを
 * 作成して、/proc/driver/sample/sample_data エントリを作る。
 */
#define PARENT_DRV_PROCNAME NULL
#define DRV_PROCNAME "driver/sample"
#define PROCNAME "sample_data"

#define DUMMY_BUFSIZ 4096
static char dummy[DUMMY_BUFSIZ];

static struct proc_dir_entry *sample_proc = NULL;

static int sample_proc_show(struct seq_file *m, void *v)
{
	seq_write(m, dummy, DUMMY_BUFSIZ);
    return 0;
}

static int sample_proc_open(struct inode *inode, struct file *file)
{
    return single_open(file, sample_proc_show, NULL);
}

static const struct file_operations sample_proc_fops = {
    .owner      = THIS_MODULE,
    .open       = sample_proc_open,
    .read       = seq_read,
    .llseek     = seq_lseek,
    .release    = single_release,
    .write      = NULL,
};

static int sample_init(void)
{
	struct proc_dir_entry *entry;
	int i;

	/* /proc 配下にディレクトリを作る */
	sample_proc = proc_mkdir(DRV_PROCNAME, PARENT_DRV_PROCNAME);
	if (sample_proc == NULL) {
		printk(KERN_ERR "Unable to create `%s' proc directory", DRV_PROCNAME);
		goto error;
	}

	/* ディレクトリ配下にエントリを作る */
	entry = proc_create(PROCNAME, S_IRUGO | S_IWUSR, sample_proc, &sample_proc_fops);
	if (entry == NULL) {
		printk(KERN_WARNING "sample: unable to create /proc entry\n");
		goto error;
	}

	for (i = 0 ; i < DUMMY_BUFSIZ ; i++) {
		dummy[i] = 'A' + (i % 26);
	}

	printk("driver loaded\n");
		
	return 0;

error:
	if (sample_proc) {
		remove_proc_entry(DRV_PROCNAME, PARENT_DRV_PROCNAME);
		sample_proc = NULL;
	}
	return -EIO;
}

static void sample_exit(void)
{
	if (sample_proc) {
		/* 不要になったエントリおよびディレクトリを削除する */
		remove_proc_entry(PROCNAME, sample_proc);
		remove_proc_entry(DRV_PROCNAME, PARENT_DRV_PROCNAME);
		sample_proc = NULL;
	}

	printk(KERN_ALERT "driver unloaded\n");
}

module_init(sample_init);
module_exit(sample_exit);

