/*
 * main.c  - mmap sample driver
 * 
 * For Fedora16 alpha(Linux 3.0) 
 * (2011.8.27 Yutaka Hirata)
 * For Fedora25 (Linux 4.8.6)
 * (2016.12.14 Yutaka Hirata)
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <asm/current.h>
#include <asm/uaccess.h>

MODULE_LICENSE("Dual BSD/GPL");

/* メモリマップ先の物理アドレス(アーキテクチャ依存)。
 * ここでは、x86 の APICBASE を指定している。
 */
#define MMAP_ADDR	0xFEE00000

#define DRIVER_NAME "sample"

static unsigned int sample_major = 260;
module_param(sample_major, uint, 0);

/* メモリマップ用ハンドラ */
static int sample_mmap(struct file *file, struct vm_area_struct *vma)
{
	unsigned long addr = MMAP_ADDR;

	printk("%s: mmap open\n", __func__);

	if ((vma->vm_end - vma->vm_start) != PAGE_SIZE || vma->vm_pgoff) {
		printk("%s: mmap error1\n", __func__);
		return -EINVAL;
	}

	if (addr & (PAGE_SIZE - 1)) {
		printk("%s: mmap error2\n", __func__);
		return -ENOSYS;
	}

	vma->vm_flags |= VM_IO;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	if (io_remap_pfn_range(vma, vma->vm_start, addr >> PAGE_SHIFT, 
				PAGE_SIZE, vma->vm_page_prot)) {
		printk("%s: mmap error3\n", __func__);
		return -EAGAIN;
	}

	return 0;
}

static int sample_open(struct inode *inode, struct file *file)
{
	return 0;     /* success */
}

static int sample_close(struct inode *inode, struct file *file)
{
	return 0;     /* success */
}

static struct file_operations sample_fops = {
	.open = sample_open,
	.release = sample_close,
	.mmap = sample_mmap,
};

static int sample_init(void)
{
	int major;
	int ret = 0;

	major = register_chrdev(sample_major, DRIVER_NAME, &sample_fops);
	if ((sample_major > 0 && major != 0) ||        /* static allocation */
			(sample_major == 0 && major < 0) ||    /* dynamic allocation */
			major < 0) {                           /* else */
		printk("%s driver registration error\n", DRIVER_NAME);
		ret = major;
		goto error;
	}
	if (sample_major == 0) {   /* dynamic allocation */
		sample_major = major;
	}

	printk("%s driver(major %d) installed.\n", DRIVER_NAME, sample_major);

error:
	return (ret);
}

static void sample_exit(void)
{
	unregister_chrdev(sample_major, DRIVER_NAME);

	printk("%s driver removed.\n", DRIVER_NAME);

}

module_init(sample_init);
module_exit(sample_exit);

